from pandas import DataFrame
from numpy import ndarray

from typing import TypedDict, Literal, NotRequired, Protocol, Optional, Any
from scipy.sparse import (
    bsr_matrix,
    coo_matrix,
    csc_matrix,
    csr_matrix,
    dia_matrix,
    dok_matrix,
    lil_matrix,
)


class ScikitModel(Protocol):
    def predict_proba(self, X) -> Optional[Any]: ...


class ViewMeta(TypedDict):
    mode: NotRequired[Literal["public", "protected", "private"]]
    view_id: NotRequired[str]
    dataset_name: NotRequired[str]


class ValuateModelOutput(TypedDict):
    view_id: str | None
    view_url: str


class Credentials(TypedDict):
    email: str
    password: str


class AWSVariables(TypedDict):
    api_url: NotRequired[str]
    app_client_id: NotRequired[str]
    identity_pool_id: NotRequired[str]
    user_pool_id: NotRequired[str]
    bucket_name: NotRequired[str]
    base_url: NotRequired[str]
    validation_api_url: NotRequired[str]


class ColumnNames(TypedDict):
    dataset_column_names: NotRequired[list[str]]
    score_column_names: NotRequired[list[str]]
    dependent_variable_name: NotRequired[str]


class GetScoreColumnNamesParams(TypedDict):
    model_name: str
    scores: list[str]
    column_names: list[str]


type Data = DataFrame | list[str | int | float] | list[
    list[str | int | float]
    # Sparse matrix types Ref: https://docs.scipy.org/doc/scipy/reference/sparse.html#sparse-matrix-classes
] | ndarray | bsr_matrix | coo_matrix | csc_matrix | csr_matrix | dia_matrix | dok_matrix | lil_matrix



from aiohttp import ClientSession
from numpy import ndarray
from pandas import DataFrame, RangeIndex
from json import dump
from scipy.sparse import spmatrix
from gooder_ai.types import GetScoreColumnNamesParams, Data
import logging


async def validate_config(validation_url: str, config: dict) -> dict:
    async with ClientSession() as session:
        output = None
        async with session.post(validation_url, json=config) as response:
            if response.status == 200:
                output = await response.json()
            else:
                error = await response.text()
                logging.error(f"Validation failed with {response.status}: {error}")
                raise Exception(f"Validation failed with {response.status}: {error}")
        return output


def get_dataset_column_names(dataset: ndarray, column_names: list[str]) -> list[str]:
    column_count = (
        len(dataset) if dataset.ndim == 1 else len(dataset[0])
    )  # ndim checks the dimensions of numpy array
    # If column names mismatch, log warning
    if len(column_names) != column_count:
        logging.warning(
            f"Mismatch: Dataset has {column_count} columns, but {len(column_names)} column names were provided."
        )

    # Ensure dataset_column_names has enough names, filling missing ones
    output = column_names[:column_count] + [
        f"column-{i+1}" for i in range(len(column_names), column_count)
    ]

    return output


# TODO: Verify if need to handle multi output classifiers.
def get_score_column_names(params: GetScoreColumnNamesParams) -> list[str]:
    column_names = params["column_names"]
    scores = params["scores"]
    model_name = params["model_name"]

    column_names_count = len(column_names)
    scores_count = len(scores)

    # Log warning if there's a mismatch
    if column_names_count != scores_count:
        logging.warning(
            f"Mismatch: Scores has {scores_count} possible values, but {column_names_count} column names were provided."
        )

    # Ensure output has enough names, filling missing ones
    output = [f"{model_name}-{name}" for name in column_names[:scores_count]] + [
        f"{model_name}-score-class-{column_name}"
        for column_name in scores[column_names_count:]
    ]

    return output


def get_scorer_functions(models_count: int, scorers: list[str]) -> list[str]:
    if models_count != len(scorers):
        logging.warning(
            f"Mismatch:  {models_count} Models, but {len(scorers)} scoring functions were provided."
        )

    output = (
        scorers[:models_count]
        + ["predict_proba" for _ in range(len(scorers), models_count)]
        if models_count != len(scorers)
        else scorers
    )

    return output


# Add column names to dataframe
def transform_dataframe(data: DataFrame, column_names: list[str]) -> DataFrame:
    transformed_dataframe = data.copy()

    # Get existing column names
    default_column_names = transformed_dataframe.columns
    column_count = transformed_dataframe.shape[1]
    final_column_names = []

    if len(column_names) != column_count:
        logging.warning(
            f"Mismatch: DataFrame has {column_count} columns, but {len(column_names)} column names were provided."
        )

    if len(column_names) == column_count:
        # Priority 1: Use provided column names
        final_column_names = column_names
    elif isinstance(default_column_names, RangeIndex) == False:
        # Priority 2: Use existing column names if not RangeIndex
        final_column_names = default_column_names
    else:
        # Priority 3: Generate default names if no names available
        final_column_names = column_names[:column_count] + [
            f"column-{i+1}" for i in range(len(column_names), column_count)
        ]

    # Assign final column names
    transformed_dataframe.columns = final_column_names

    return transformed_dataframe


def transform_unstructured_numpy_array(
    data: ndarray, column_names: list[str]
) -> DataFrame:
    logging.warning("Unstructured numpy array does not have any column name")
    numpy_array = data.copy()
    final_column_names = []

    if numpy_array.ndim == 1:
        if len(column_names) == 0:
            logging.warning(
                "Mismatch: Numpy array has 1 column, but 0 column names were provided."
            )
        final_column_names = (
            [column_names[0]] if len(column_names) > 0 else ["column-1"]
        )
    elif numpy_array.ndim == 2:
        _, column_count = data.shape
        if len(column_names) != column_count:
            logging.warning(
                f"Mismatch: Numpy array has {column_count} columns, but {len(column_names)} column names were provided."
            )
        final_column_names = column_names[:column_count] + [
            f"column-{i+1}" for i in range(len(column_names), column_count)
        ]
    else:
        raise ValueError("Input data must be a 1D or 2D NumPy array.")
    output = DataFrame(numpy_array, columns=final_column_names)
    return output


def transform_structured_numpy_array(
    data: ndarray, column_names: list[str]
) -> DataFrame:
    numpy_array = data.copy()
    final_column_names = []
    default_column_names = list(data.dtype.names)

    if numpy_array.ndim == 1:
        if len(column_names) == 1:
            final_column_names = [column_names[0]]
        else:
            logging.warning(
                "Mismatch: Numpy array has 1 column, but 0 column names were provided."
            )
            final_column_names = (
                [default_column_names[0]]
                if len(default_column_names) > 0
                else ["column-1"]
            )

    elif numpy_array.ndim == 2:
        _, column_count = data.shape
        if len(column_names) != column_count:
            logging.warning(
                f"Mismatch: Numpy array has {column_count} columns, but {len(column_names)} column names were provided."
            )
        if len(column_names) == column_count:
            final_column_names = column_names
        else:
            final_column_names = column_names[:column_count] + [
                f"column-{i+1}" for i in range(len(column_names), column_count)
            ]
    else:
        raise ValueError("Input data must be a 1D or 2D NumPy array.")
    output = DataFrame(numpy_array, columns=final_column_names)
    return output


def transform_list(
    data: (
        list[str]
        | list[int]
        | list[float]
        | list[list[str]]
        | list[list[int]]
        | list[list[float]]
    ),
    column_names: list[str],
) -> DataFrame:
    logging.warning("List does not have any column name.")
    data_list = data.copy()
    final_column_names = []

    # 1D list handling.
    if isinstance(data_list[0], (str, int, float)):
        data_list = [[value] for value in data_list]
        final_column_names = column_names[:1] if column_names else ["column-1"]

    elif isinstance(data_list[0], list):
        column_count = len(data_list[0])
        if len(column_names) != column_count:
            logging.warning(
                f"Mismatch: DataFrame has {column_count} columns, but {len(column_names)} column names were provided."
            )
        final_column_names = column_names[:column_count] + [
            f"column-{i+1}" for i in range(len(column_names), column_count)
        ]

    else:
        raise ValueError("Input data must be a 1D or 2D list.")

    # Create DataFrame
    output = DataFrame(data_list, columns=final_column_names)
    return output


def get_transformed_data(data: Data, column_names: list[str]) -> DataFrame:
    output = DataFrame()
    if isinstance(data, DataFrame):
        output = transform_dataframe(data, column_names)
    elif isinstance(data, ndarray) and data.dtype.names is None:
        output = transform_unstructured_numpy_array(data, column_names)
    elif isinstance(data, ndarray) and data.dtype.names is not None:
        transform_structured_numpy_array(data, column_names)
    elif isinstance(data, list):
        output = transform_list(data, column_names)
    elif isinstance(data, spmatrix):
        """
        scipy has multiple types of matrix all of them are extending spmatrix
        Ref: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.spmatrix.html#spmatrix
        """
        numpy_array = data.toarray()
        output = transform_unstructured_numpy_array(numpy_array, column_names)

    if isinstance(output.columns, RangeIndex):
        raise ValueError("Invalid columns")
    return output


def save_config(config: dict, file_path: str):
    with open(file_path, "w") as file:
        # dump automatically writes to file
        dump(
            config, file, indent=4
        )  # Ref: https://docs.python.org/3/library/json.html#basic-usage



import boto3
from gooder_ai.auth.types import AuthenticateParams

Cognito_Client = boto3.client(
    "cognito-idp"
)  # Initialize a boto3 client for Cognito Identity Provider to manage authentication flows
Cognito_Identity_Client = boto3.client(
    "cognito-identity"
)  # Initialize a boto3 client for Cognito Identity to manage access key and secret access key


def authenticate(input: AuthenticateParams) -> dict:
    email = input["email"]
    password = input["password"]
    app_client_id = input["app_client_id"]
    identity_pool_id = input["identity_pool_id"]
    user_pool_id = input["user_pool_id"]

    # Initiate authentication Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cognito-idp/client/initiate_auth.html#initiate-auth
    # cognito_client_response consists of access_token, id_token, refresh_token.
    cognito_client_response = Cognito_Client.initiate_auth(
        AuthFlow="USER_PASSWORD_AUTH",
        ClientId=app_client_id,
        AuthParameters={"USERNAME": email, "PASSWORD": password},
    )

    # Using id_token to get the client_id. Ref: https://stackoverflow.com/a/62789387
    cognito_identity_response = Cognito_Identity_Client.get_id(
        IdentityPoolId=identity_pool_id
    )
    # Using cognito identity id to get access_key, secret_access_key which will be used to upload files to s3. Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cognito-identity/client/get_credentials_for_identity.html#get-credentials-for-identity
    cognito_credentials_response = Cognito_Identity_Client.get_credentials_for_identity(
        IdentityId=cognito_identity_response["IdentityId"],
        Logins={
            f"cognito-idp.us-east-1.amazonaws.com/{user_pool_id}": cognito_client_response[
                "AuthenticationResult"
            ][
                "IdToken"
            ]
        },
    )

    credentials = {
        "cognito_client_response": cognito_client_response,
        "cognito_identity_response": cognito_identity_response,
        "cognito_credentials": cognito_credentials_response,
    }

    return credentials



# The following configuration values are environment-specific and need to be updated when switching to a new environment.
# For example, we are currently on environment "seventeen," and when transitioning to environment "eighteen,"
# these values should be updated accordingly.

API_URL = "https://ojykj5ys3rflbprrjwiyi524oy.appsync-api.us-east-1.amazonaws.com/graphql"  # GraphQL API URL
App_Client_ID = "6hmbs2hm1dbpantdbpfnokina"  # App_Client_ID is the client ID for an app in a Cognito User Pool
Identity_Pool_ID = "us-east-1:9682a196-3376-4d36-b5d8-d46862f47d7b"  # Identity_Pool_ID is the ID for an Amazon Cognito Identity Pool
User_Pool_ID = "us-east-1_wzrj8LRkh"  # UserPool_ID is the ID for the Cognito User Pool that manages user authentication
Bucket_Name = "mlviz1049e7ed428840c3bb9773b134209edd111800-seventeen"  # S3 bucket name storing datasets and configs
Validation_API_URL = "https://htknl8tmxi.execute-api.us-east-1.amazonaws.com/seventeen/validate"  # Validation API URL
# This is the URL for gooder ai application
Base_URL = "https://latest.gooder.ai"
AWS_Region_Name = "us-east-1"



from datetime import datetime
from uuid import uuid4
from webbrowser import open_new
from gooder_ai import globals
from gooder_ai.s3 import upload_files
from gooder_ai.view import execute_graphql_query, ExecuteGraphQLParams
from gooder_ai.auth import authenticate
from gooder_ai.utils import (
    validate_config,
    get_transformed_data,
    get_score_column_names,
    get_scorer_functions,
    save_config,
)
from gooder_ai.types import (
    ViewMeta,
    ValuateModelOutput,
    ScikitModel,
    Credentials,
    AWSVariables,
    ColumnNames,
    Data,
)
from asyncio import to_thread
from pandas import DataFrame, concat
import logging
import os


async def valuate_model(
    models: list[ScikitModel],
    x_data: Data,
    y: Data,
    auth_credentials: Credentials,
    config: dict,
    view_meta: ViewMeta,
    scorer_names: list[str] = [],
    column_names: ColumnNames = {},
    filtered_columns: list[str] = [],
    aws_variables: AWSVariables = {},
    upload_data_to_gooder: bool = True,
    upload_config_to_gooder: bool = True,
) -> ValuateModelOutput:
    logging.info("Model valuation started.")
    email = auth_credentials["email"]
    password = auth_credentials["password"]
    mode = view_meta.get("mode", "private")
    view_id = view_meta.get("view_id", None)
    dataset_name = view_meta.get("dataset_name", f"{datetime.now().timestamp()}")

    # AWS Global Variables
    api_url = aws_variables.get("api_url", globals.API_URL)
    app_client_id = aws_variables.get("app_client_id", globals.App_Client_ID)
    identity_pool_id = aws_variables.get("identity_pool_id", globals.Identity_Pool_ID)
    user_pool_id = aws_variables.get("user_pool_id", globals.User_Pool_ID)
    bucket_name = aws_variables.get("bucket_name", globals.Bucket_Name)
    base_url = aws_variables.get("base_url", globals.Base_URL)
    validation_api_url = aws_variables.get(
        "validation_api_url", globals.Validation_API_URL
    )

    transformed_x_data = get_transformed_data(
        x_data, column_names.get("dataset_column_names", [])
    )

    filtered_x_data = (
        transformed_x_data.filter(items=filtered_columns)
        if len(filtered_columns) > 0
        else transformed_x_data
    )

    transformed_y_data = get_transformed_data(
        y, [column_names.get("dependent_variable_name", "dependent_variable")]
    )
    combined_dataframe = concat([filtered_x_data, transformed_y_data], axis=1)

    scorer_functions = get_scorer_functions(len(models), scorer_names)

    for index, model in enumerate(models):
        scorer = getattr(model, scorer_functions[index])
        if scorer is None:
            logging.error(
                f"Failed: Input model instance doesn't have a '{scorer_functions[index]}' method to score the model performance."
            )
            raise Exception(
                f"Failed: Input model instance doesn't have a '{scorer_functions[index]}' method to score the model performance."
            )
        model_classes = getattr(model, "classes_", None)
        if model_classes is None:
            logging.error(
                "Failed: Input model instance doesn't have classes to classify the target variable"
            )
            raise Exception(
                "Failed: Input model instance doesn't have classes to classify the target variable."
            )
        score_column_names = get_score_column_names(
            {
                "column_names": column_names.get("score_column_names", []),
                "scores": model_classes,
                "model_name": f"model-{index}",
            }
        )
        combined_dataframe = concat(
            [
                combined_dataframe,
                DataFrame(scorer(x_data), columns=score_column_names),
            ],
            axis=1,
        )

    logging.info("Started: Validating config as per the Gooder AI schema.")
    parsed_config = await validate_config(validation_api_url, config)

    if parsed_config["success"] == False:
        logging.error("Failed: Validating config as per the Gooder AI schema.")
        raise Exception("Invalid configuration", parsed_config["error"])
    else:
        logging.info("Success: Validating config as per the Gooder AI schema.")

    logging.info("Started: Authenticating for the Gooder AI platform.")
    credentials = authenticate(
        {
            "email": email,
            "password": password,
            "app_client_id": app_client_id,
            "identity_pool_id": identity_pool_id,
            "user_pool_id": user_pool_id,
        }
    )
    logging.info("Success: Authenticating for the Gooder AI platform.")

    token = credentials["cognito_client_response"]["AuthenticationResult"][
        "AccessToken"
    ]
    aws_access_key_id = credentials["cognito_credentials"]["Credentials"]["AccessKeyId"]
    aws_secret_access_key = credentials["cognito_credentials"]["Credentials"][
        "SecretKey"
    ]
    aws_session_token = credentials["cognito_credentials"]["Credentials"][
        "SessionToken"
    ]
    identity_id = credentials["cognito_credentials"]["IdentityId"]

    parsed_config["data"][
        "datasetID"
    ] = f"{dataset_name}.csv/Sheet1"  # override datasetID of config to match with dataset.

    if upload_data_to_gooder == True:
        logging.info("Started: Uploading dataset to the Gooder AI platform.")

    if upload_config_to_gooder == True:
        logging.info("Started: Uploading config to the Gooder AI platform.")

    path_dictionary = await upload_files(
        {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "aws_session_token": aws_session_token,
            "identity_id": identity_id,
            "data": combined_dataframe,
            "config": parsed_config["data"],
            "file_name": dataset_name,
            "mode": mode,
            "bucket_name": bucket_name,
            "upload_data_to_gooder": upload_data_to_gooder,
            "upload_config_to_gooder": upload_config_to_gooder,
        }
    )
    csv_path = path_dictionary["csv_path"]
    config_path = path_dictionary["config_path"]

    if csv_path is None and upload_data_to_gooder == True:
        logging.error("Failed: Uploading dataset to the Gooder AI platform.")
        raise Exception("Failed to upload dataset")
    elif upload_data_to_gooder == False:
        local_dataset_filename = f"{dataset_name}.csv"
        # Ref: https://docs.python.org/3/library/asyncio-task.html#running-in-threads
        # TODO: Do we need to_thread ?
        await to_thread(
            lambda: combined_dataframe.to_csv(local_dataset_filename, index=False)
        )
        logging.info(f"Dataset saved as {os.path.abspath(local_dataset_filename)}.")

    if config_path is None and upload_config_to_gooder == True:
        logging.error("Failed: Uploading config to the Gooder AI platform.")
        raise Exception("Failed to upload config")
    elif upload_config_to_gooder == False:
        local_config_filename = (
            f"{dataset_name}.json"  # JSON file name is same as dataset_name file name
        )
        # Ref: https://docs.python.org/3/library/asyncio-task.html#running-in-threads
        # TODO: Do we need to_thread ?
        await to_thread(lambda: save_config(config, local_config_filename))
        logging.info(f"Config saved as {os.path.abspath(local_config_filename)}.")

    mutation_type = (
        "updateSharedView" if isinstance(view_id, str) else "createSharedView"
    )
    view_params: ExecuteGraphQLParams = {
        "api_url": api_url,
        "token": token,
        "mutation": mutation_type,
        "variables": {
            "input": {
                "configPath": config_path,
                "datasetPath": csv_path,
                "id": view_id if isinstance(view_id, str) else f"{uuid4()}",
            }
        },
    }

    id: str | None = None
    if upload_config_to_gooder == True or upload_data_to_gooder == True:
        view = await execute_graphql_query(view_params)
        id = view["data"][mutation_type]["id"]
        message = (
            f"View with ID {id} has been successfully updated using the provided view ID: {view_id}."
            if mutation_type == "updateSharedView"
            else f"A new view has been created successfully. Your view ID is {id}. Please save it for future reference and reuse."
        )
        logging.info(message)
        logging.info("Model valuation can be continued on the Gooder AI platform now.")

    view_url = base_url if id is None else f"{base_url}?view={id}"
    open_new(view_url)
    return {"view_id": id, "view_url": view_url}
